#define RUBY_VERSION "1.8.0"
#define RUBY_RELEASE_DATE "2003-08-04"
#define RUBY_VERSION_CODE 180
#define RUBY_RELEASE_CODE 20030804

#define RUBY_VERSION_MAJOR 1
#define RUBY_VERSION_MINOR 8
#define RUBY_VERSION_TEENY 0
#define RUBY_RELEASE_YEAR 2003
#define RUBY_RELEASE_MONTH 8
#define RUBY_RELEASE_DAY 4
