module REXML
	module Encoding
		def to_utf_16 content
			array_utf8 = content.unpack("U*")
			array_enc = []
			array_utf8.each do |num|
				if ((num>>16) > 0)
					array_enc << 0
					array_enc << ??
				else
					array_enc << (num >> 8)
					array_enc << (num & 0xFF)
				end
			end
			array_enc.pack('C*')
		end

		def from_utf_16(str)
			array_enc=str.unpack('C*')
			array_utf8 = []
			2.step(arrayEnc.size-1, 2){|i| 
				array_utf8 << (array_enc.at(i+1) + array_enc.at(i)*0x100)
			}
			array_utf8.pack('U*')
		end
	end
end
