# REXML is an XML parser for Ruby, in Ruby.
# 
# URL: http://www.germane-software.com/software/rexml
# Author: Sean Russell <ser@germane-software.com>
# Version: 2.5.6
# Date: +2003/054



# 
# Short Description:
# Why did I write REXML?   At the time of this writing, there were already
# two XML parsers for Ruby. The first is a Ruby binding to a native XML
# parser.  This is a fast parser, using proven technology. However,
# it isn't very portable. The second is a native Ruby implementation, but
# I didn't like its API very much.  I wrote REXML for myself, so that I'd
# have an XML parser that had an intuitive API.
#
# API documentation can be downloaded from the REXML home page, or can
# be accessed online at http://www.germane-software.com/software/rexml_doc
# A tutorial is available in docs/tutorial.html
module REXML
	Copyright = "Copyright #{Time.now.year} Sean Russell <ser@germane-software.com>"
	Date = "+2003/110"
	Version = "2.7.1"
end
