alias $ERROR_INFO              $!
alias $ERROR_POSITION          $@
alias $LOADED_FEATURES         $"
alias $FS                      $;
alias $FIELD_SEPARATOR         $;
alias $OFS                     $,
alias $OUTPUT_FIELD_SEPARATOR  $,
alias $RS                      $/
alias $INPUT_RECORD_SEPARATOR  $/
alias $ORS                     $\
alias $OUTPUT_RECORD_SEPARATOR $\
alias $INPUT_LINE_NUMBER       $.
alias $NR                      $.
alias $LAST_READ_LINE          $_
alias $DEFAULT_OUTPUT          $>
alias $DEFAULT_INPUT           $<
alias $PID                     $$
alias $PROCESS_ID              $$
alias $CHILD_STATUS            $?
alias $LAST_MATCH_INFO         $~
alias $IGNORECASE              $=
alias $PROGRAM_NAME            $0
alias $ARGV                    $*
alias $MATCH                   $&
alias $PREMATCH                $`
alias $POSTMATCH               $'
alias $LAST_PAREN_MATCH        $+
