EXPORTS
test_alloc_test_struct
test_append
test_arylen
test_c2i
test_call_func1
test_callback1
test_close
test_d2f
test_f2d
test_fill_test_struct
test_fill_test_union
test_gets
test_i2c
test_init
test_isucc
test_lcc
test_lsucc
test_open
test_strcat
test_strlen
test_succ
test_data_init
test_data_add
test_data_aref
test_set_long_value
test_get_long_value
internal_long_value
